/*
 * IRKGAUSSMEX-Interface by C. Ludwig
 * Version: $Id$ */
#define IRKGAUSSMexVersion "14. Dec. 2006"
/*
 *
 * Questions, requests, problems, suggestions,
 * remarks and bugs to
 * Ludwig_C@gmx.de
 */
#include <math.h>
#include <string.h>
#include "mex.h"
#include "options.h"
#include "irkgaussMex.h"

static SOptionSettings optSet;
static SParameterIO paramIO;
static SParameterGlobal paramGlobal;
static SParameterIRKGauss paramIRK;
static SParameterRightSide paramRightSide;

/*
static void debug (const char *msg) {
  mxArray *plhs[1];

  mexPrintf(msg);
  mexPrintf("\n");
  plhs[0]=NULL;
  mexCallMATLAB(0,plhs,0,NULL,"drawnow"); 
  if (plhs[0]!=NULL) {
    mxDestroyArray(plhs[0]);plhs[0]=NULL;
  }
}
*/

static char ismxArrayString (const mxArray *arr) {
  if (arr==NULL) return (char)0;
  if (mxIsChar(arr)) return (char)1;
  return (char)0;
}

static char ismxArrayFunction (const mxArray *arr) {
  if (arr==NULL) return (char)0;
  return (mxGetClassID(arr)==mxFUNCTION_CLASS)?(char)1:(char)0;
}

static char ismxArrayInline (const mxArray *arr) {
  if (arr==NULL) return (char)0;
  return (strcmp(mxGetClassName(arr),"inline")==0)?(char)1:(char)0;
}

static void initVars (void) {
  /* IO parameters */
  paramIO.opt=NULL;paramIO.optCreated=0;

  /* Option settings */
  optSet.warnMiss=0;optSet.warnType=1;optSet.warnSize=1;

  /* global parameters */
  paramGlobal.tPointer=NULL;
  paramGlobal.tGrid=NULL;paramGlobal.ptGrid=NULL;
  paramGlobal.xGrid=NULL;paramGlobal.pxGrid=NULL;

  /* parameters for IRK Gauss */
  paramIRK.IOUT=0;
  paramIRK.x=NULL;

  /* parameters for right side */
  paramRightSide.rightSideFcn=NULL;
  paramRightSide.rightSideFcnH=NULL;
  paramRightSide.tArg=NULL;paramRightSide.xArg=NULL;
}

static void doneVars (void) {
  /* IO parameters */
  if ((paramIO.optCreated) && (paramIO.opt!=NULL)) {
    mxDestroyArray((mxArray*)paramIO.opt);paramIO.opt=NULL;
  }
  
  /* global parameters */
  /* tPointer belongs to the caller */
  paramGlobal.tPointer=NULL;
  /* tGrid,xGrid are return values => don't destroy */
  paramGlobal.ptGrid=NULL;paramGlobal.pxGrid=NULL;

  /* parameters for IRK Gauss */
  if (paramIRK.x!=NULL) {
    mxFree(paramIRK.x);paramIRK.x=NULL;
  }

  /* parameters for right side */
  if (paramRightSide.rightSideFcn!=NULL) {
    mxFree(paramRightSide.rightSideFcn);paramRightSide.rightSideFcn=NULL;
  }
  paramRightSide.rightSideFcnH=NULL; /* belongs to caller */
  if (paramRightSide.tArg!=NULL) {
    mxDestroyArray(paramRightSide.tArg);paramRightSide.tArg=NULL;
  }
  if (paramRightSide.xArg!=NULL) {
    mxDestroyArray(paramRightSide.xArg);paramRightSide.xArg=NULL;
  }
}

static void stopMexFunction (int errNo,
  int i1, int i2, int i3, double d1) {

  char *msg;
  mexPrintf("Error (%i) [Version: %s]:\n",errNo,IRKGAUSSMexVersion);
  switch (errNo) {
    #include "errors.c"
    default: msg="unknown error number";break;
  }

  doneVars();
  mexErrMsgTxt(msg);
}

static void checkNumberOfArgs (int nlhs, int nrhs) {
  if (nlhs!=2) stopMexFunction(1,nlhs,0,0,0);
  if ((nrhs!=3) && (nrhs!=4)) stopMexFunction(2,nrhs,0,0,0);
}

static void processArgs (int nrhs, const mxArray *prhs[]) {
  int buflen;
  double *dpointer;

  /* 1st arg: right side */
  if (ismxArrayString(prhs[0])) {
    if ((mxGetNumberOfDimensions(prhs[0])!=2) || (mxGetM(prhs[0])!=1))
      stopMexFunction(3,mxGetNumberOfDimensions(prhs[0]),mxGetM(prhs[0]),
      0,0);
    buflen=mxGetN(prhs[0])*sizeof(mxChar)+1;
    paramRightSide.rightSideFcn=(char*)mxMalloc(buflen);
    mxGetString(prhs[0],paramRightSide.rightSideFcn,buflen);
    paramRightSide.rightSideFcnH=prhs[0];
  } else 
  if ( (ismxArrayFunction(prhs[0])) || (ismxArrayInline(prhs[0])) ) {
    if ((mxGetNumberOfDimensions(prhs[0])!=2) || (mxGetM(prhs[0])!=1) ||
        (mxGetN(prhs[0])!=1))
      stopMexFunction(4,mxGetNumberOfDimensions(prhs[0]),
        mxGetM(prhs[0]),mxGetN(prhs[0]),0);
    paramRightSide.rightSideFcn=NULL; 
    paramRightSide.rightSideFcnH=prhs[0];
  } else {
    stopMexFunction(5,0,0,0,0);
  }

  /* 2nd arg: row-vector containing time-values */
  if (!mxIsDouble(prhs[1]) || mxIsSparse(prhs[1])) stopMexFunction(6,0,0,0,0);
  if ((mxGetNumberOfDimensions(prhs[1])!=2) || (mxGetM(prhs[1])!=1) ||
      (mxGetN(prhs[1])!=2) )
    stopMexFunction(7,mxGetNumberOfDimensions(prhs[1]),mxGetM(prhs[1]),0,0);
  dpointer=mxGetPr(prhs[1]);paramGlobal.tPointer=dpointer;
  paramIRK.tau=dpointer[0];paramIRK.tauEnd=dpointer[1];
  /* no checks for tau<tauEnd !!! */

  /* 3rd arg: start vector */
  if (!mxIsDouble(prhs[2]) || mxIsSparse(prhs[2])) stopMexFunction(8,0,0,0,0);
  if ((mxGetNumberOfDimensions(prhs[2])!=2) || (mxGetN(prhs[2])!=1))
    stopMexFunction(9,mxGetNumberOfDimensions(prhs[2]),mxGetN(prhs[2]),0,0);
  paramGlobal.d=mxGetM(prhs[2]);
  if (paramGlobal.d<1) stopMexFunction(10,0,0,0,0);
  dpointer=mxGetPr(prhs[2]);
  paramIRK.x=(double*)mxMalloc(paramGlobal.d*sizeof(double));
  memcpy(paramIRK.x,dpointer,paramGlobal.d*sizeof(double));
  /* little remark: A COPY of the startvector is made, because
     it will be passed to Fortran code. To be sure, that
     it will not be changed there, a copy will be passed.
     x0 belongs to the caller and it is not allowed 
     (due to Matlab-Contract) to change it. */

  /* 4th arg: struct with options */
  if (nrhs==4) {
    if (!mxIsStruct(prhs[3])) stopMexFunction(11,0,0,0,0);
    paramIO.opt=prhs[3];paramIO.optCreated=0;
  } else {
    paramIO.opt=mxCreateStructMatrix(1,1,0,NULL);
    paramIO.optCreated=1;
  }
}

static void extractOptionSettings (void) {
  optSet.warnMiss=opt_getIntFromOpt(paramIO.opt,&optSet,OPT_WARNMISS,0);
  optSet.warnType=opt_getIntFromOpt(paramIO.opt,&optSet,OPT_WARNTYPE,1);
  optSet.warnSize=opt_getIntFromOpt(paramIO.opt,&optSet,OPT_WARNSIZE,1);
}

static void extractNStepAndMethod (void) {
  paramIRK.nstep=opt_getIntFromOpt(paramIO.opt,&optSet,OPT_NSTEP,10);
  paramIRK.method=opt_getIntFromOpt(paramIO.opt,&optSet,OPT_METHOD,3);

  switch (paramIRK.method) {
    case 2: case 3: case 4: case 6: break;
    default: stopMexFunction(12,paramIRK.method,0,0,0);
  }
}

static void extractGlobalOptions (void) {
  paramGlobal.funcCallMethod=opt_getIntFromOpt(paramIO.opt,&optSet,
    OPT_FUNCCALLMETHOD,1);
  switch (paramGlobal.funcCallMethod) {
    case 0: /* use mexCallMATLAB direct */
      if (paramRightSide.rightSideFcn==NULL) stopMexFunction(13,0,0,0,0);
      break;
    case 1: /* use mexCallMATLAB to call feval */
      break;
    default:
      stopMexFunction(14,0,0,0,0);
      break;
  }
}

static void extractOptions (void) {
  extractOptionSettings();
  extractNStepAndMethod();
  extractGlobalOptions();
}

static void prepareHelpMxArrays (void) {
  int d;

  d=paramGlobal.d;

  paramRightSide.tArg=mxCreateDoubleMatrix(1,1,mxREAL);
  paramRightSide.xArg=mxCreateDoubleMatrix(d,1,mxREAL);
  
  paramGlobal.tGrid=mxCreateDoubleMatrix(paramIRK.nstep+1,1,mxREAL);
  paramGlobal.xGrid=mxCreateDoubleMatrix(paramIRK.nstep+1,paramGlobal.d,mxREAL);
  paramGlobal.ptGrid=mxGetPr(paramGlobal.tGrid);
  paramGlobal.pxGrid=mxGetPr(paramGlobal.xGrid);
}

void IRKGaussRightSideFunc (int *n, double *t, double *x, double *f) {
  int d,offset;
  mxArray *rhs[3];
  mxArray *lhs[1];
  
  d=*n;lhs[0]=NULL;
  /* Call by value */
  /* Use always the SAME Matlab tArg and xArg */
  /* IMPORTANT NOTICE (if the right side is also a MEX-File)
     the right side must not "garble" the passed mxArrays:
     the size must not be changed and the memory must not
     be freed. 
     Hence: the right side has to take care, that the
     passed mxArrays have the same size and enough memory
     when returning. The values in the memory(-block)
     may be overwritten.
     If the right side is an m-file MATLAB obeys this
     restriction automatically. */
  offset=0;
  switch (paramGlobal.funcCallMethod) {
    case 0: break;
    case 1: 
      rhs[offset++]=(mxArray*)paramRightSide.rightSideFcnH;
      break;
    default: stopMexFunction(1001,0,0,0,0);
  }
  rhs[offset]=paramRightSide.tArg;*mxGetPr(rhs[offset])= *t;
  offset++;
  
  rhs[offset]=paramRightSide.xArg;
  memcpy(mxGetPr(rhs[offset]),x,d*sizeof(double));
  offset++;
  
  switch (paramGlobal.funcCallMethod) {
    case 0:
      mexCallMATLAB(1,lhs,offset,rhs,paramRightSide.rightSideFcn);
      break;
    case 1:
      mexCallMATLAB(1,lhs,offset,rhs,"feval");
      break;
    default: stopMexFunction(1001,0,0,0,0);break;
  }
  
  /* check return values */
  if (lhs[0]==NULL) stopMexFunction(301,0,0,0,0);
  if ((!mxIsDouble(lhs[0])) || mxIsSparse(lhs[0]) ||
      (mxGetNumberOfDimensions(lhs[0])!=2))
    stopMexFunction(302,0,0,0,0);
  if (! (((mxGetM(lhs[0])==d) && (mxGetN(lhs[0])==1)) ||
         ((mxGetM(lhs[0])==1) && (mxGetN(lhs[0])==d))) )
    stopMexFunction(303,mxGetM(lhs[0]),mxGetN(lhs[0]),0,0);
  memcpy(f,mxGetPr(lhs[0]),d*sizeof(double));
  
  mxDestroyArray(lhs[0]);lhs[0]=NULL;
}

void IRKGaussSaveStateFunc (int *n, int *iter, double *tOld, 
  double *t, double *x) {
  int i;

  paramGlobal.ptGrid[*iter]= *t;
  for (i=0; i< *n; i++) {
    paramGlobal.pxGrid[ (*iter) + i*(paramIRK.nstep+1) ] = x[i];
  }
}


static void createOutput (int nlhs, mxArray *plhs[]) {
  plhs[0]=paramGlobal.tGrid;
  plhs[1]=paramGlobal.xGrid;
}

void mexFunction (int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[]) {
  initVars();
  
  checkNumberOfArgs(nlhs,nrhs);
  processArgs(nrhs,prhs);

  extractOptions();
  prepareHelpMxArrays();

  IRKGAUSS_ (&paramGlobal.d, &IRKGaussRightSideFunc,
    &IRKGaussSaveStateFunc,
    &paramIRK.nstep, &paramIRK.tau, paramIRK.x, &paramIRK.tauEnd,
    &paramIRK.method, &paramIRK.IOUT);

  createOutput(nlhs,plhs);

  doneVars();
}

